@@ -6,6 +6,7 @@ import sqlite3
 import string
 import warnings
 from collections import OrderedDict
+from contextlib import contextmanager
 from datetime import datetime
 from datetime import timedelta
 from functools import partial
@@ -14,6 +15,7 @@ from types import MappingProxyType
 from typing import Any
 from typing import Dict
 from typing import Iterable
+from typing import Iterator
 from typing import List
 from typing import Optional
 from typing import Tuple
@@ -29,13 +31,19 @@ from ._sqlite_utils import require_version
 from ._sqlite_utils import SQLiteType
 from ._sqlite_utils import wrap_exceptions
 from ._sqlite_utils import wrap_exceptions_iter
-from ._storage import apply_filter_options
+from ._storage import apply_entry_filter_options
+from ._storage import apply_random
 from ._storage import apply_recent
+from ._storage import Storage
 from ._types import EntryFilterOptions
+from ._utils import exactly_one
 from ._utils import join_paginated_iter
+from ._utils import zero_or_one
+from .exceptions import EntryNotFoundError
 from .exceptions import InvalidSearchQueryError
 from .exceptions import SearchError
 from .exceptions import SearchNotEnabledError
+from .types import EntrySearchCounts
 from .types import EntrySearchResult
 from .types import HighlightedString
 from .types import SearchSortOrder
@@ -73,6 +81,8 @@ _T = TypeVar('_T')
 # Currently there's no way to allow users to pick a parser, and we don't want
 # to force a specific parser, so there's no point in warning.
 #
+# When changing this, also change the equivalent pytest.filterwarnings config.
+#
 # TODO: Expose BeautifulSoup(features=...) when we have a config system.
 #
 warnings.filterwarnings(
@@ -103,10 +113,40 @@ def strip_html(text: SQLiteType, features: Optional[str] = None) -> SQLiteType:
     return rv
 
 
+_QUERY_ERROR_MESSAGE_FRAGMENTS = [
+    "fts5: syntax error near",
+    "unknown special query",
+    "no such column",
+    "no such cursor",
+    "unterminated string",
+]
+
+
+@contextmanager
+def wrap_search_exceptions() -> Iterator[None]:
+    try:
+        yield
+    except sqlite3.OperationalError as e:
+        msg_lower = str(e).lower()
+
+        if 'no such table' in msg_lower:
+            raise SearchNotEnabledError()
+
+        is_query_error = any(
+            fragment in msg_lower for fragment in _QUERY_ERROR_MESSAGE_FRAGMENTS
+        )
+        if is_query_error:
+            raise InvalidSearchQueryError(message=str(e))
+
+        raise
+
+
 # When trying to fix "database is locked" errors or to optimize stuff,
 # have a look at the lessons here first:
 # https://github.com/lemon24/reader/issues/175#issuecomment-657495233
 
+# When adding a new method, add a new test_search.py::test_errors_locked test.
+
 
 class Search:
 
@@ -738,24 +778,67 @@ class Search:
         log.debug("Search.update: _insert_into_search: chunk done")
         return True
 
-    _query_error_message_fragments = [
-        "fts5: syntax error near",
-        "unknown special query",
-        "no such column",
-        "no such cursor",
-        "unterminated string",
-    ]
-
     def search_entries(
         self,
         query: str,
         now: datetime,
         filter_options: EntryFilterOptions = EntryFilterOptions(),  # noqa: B008
         sort: SearchSortOrder = 'relevant',
+        limit: Optional[int] = None,
+        starting_after: Optional[Tuple[str, str]] = None,
     ) -> Iterable[EntrySearchResult]:
-        yield from join_paginated_iter(
-            partial(self.search_entries_page, query, now, filter_options, sort),
-            self.chunk_size,
+        # TODO: dupe of at least Storage.get_entries(), maybe deduplicate
+        if sort in ('relevant', 'recent'):
+
+            last = None
+            if starting_after:
+                if sort == 'recent':
+                    assert isinstance(self.storage, Storage)
+                    last = self.storage.get_entry_last(now, sort, starting_after)
+                else:
+                    last = self.search_entry_last(query, starting_after)
+
+            rv = join_paginated_iter(
+                partial(self.search_entries_page, query, now, filter_options, sort),  # type: ignore[arg-type]
+                self.chunk_size,
+                last,
+                limit or 0,
+            )
+
+        elif sort == 'random':
+            assert not starting_after
+            it = self.search_entries_page(
+                query,
+                now,
+                filter_options,
+                sort,
+                min(limit, self.chunk_size or limit) if limit else self.chunk_size,
+            )
+            rv = (entry for entry, _ in it)
+
+        else:
+            assert False, "shouldn't get here"  # noqa: B011; # pragma: no cover
+
+        yield from rv
+
+    @wrap_exceptions(SearchError)
+    def search_entry_last(self, query: str, entry: Tuple[str, str]) -> Tuple[Any, ...]:
+        feed_url, entry_id = entry
+
+        sql_query = (
+            Query()
+            .SELECT('min(rank)', '_feed', '_id')
+            .FROM("entries_search")
+            .WHERE("entries_search MATCH :query")
+            .WHERE("_feed = :feed AND _id = :id")
+            .GROUP_BY('_feed', '_id')
+        )
+
+        context = dict(feed=feed_url, id=entry_id, query=query)
+
+        return zero_or_one(
+            self.db.execute(str(sql_query), context),
+            lambda: EntryNotFoundError(feed_url, entry_id),
         )
 
     @wrap_exceptions_iter(SearchError)
@@ -779,7 +862,6 @@ class Search:
 
         context = dict(
             query=query,
-            **filter_options._asdict(),
             **query_context,
             before_mark=before_mark,
             after_mark=after_mark,
@@ -827,25 +909,22 @@ class Search:
                 MappingProxyType(rv_content),
             )
 
-        try:
+        with wrap_search_exceptions():
             yield from paginated_query(
                 self.db, sql_query, context, value_factory, chunk_size, last
             )
 
-        except sqlite3.OperationalError as e:
-            msg_lower = str(e).lower()
-
-            if 'no such table' in msg_lower:
-                raise SearchNotEnabledError()
-
-            is_query_error = any(
-                fragment in msg_lower
-                for fragment in self._query_error_message_fragments
-            )
-            if is_query_error:
-                raise InvalidSearchQueryError(message=str(e))
-
-            raise
+    @wrap_exceptions(SearchError)
+    @wrap_search_exceptions()
+    def search_entry_counts(
+        self,
+        query: str,
+        filter_options: EntryFilterOptions = EntryFilterOptions(),  # noqa: B008
+    ) -> EntrySearchCounts:
+        sql_query, query_context = make_search_entry_counts_query(filter_options)
+        context = dict(query=query, **query_context)
+        row = exactly_one(self.db.execute(str(sql_query), context))
+        return EntrySearchCounts(*row)
 
 
 def make_search_entries_query(
@@ -886,11 +965,11 @@ def make_search_entries_query(
         .LIMIT("-1 OFFSET 0")
     )
 
-    context = apply_filter_options(search, filter_options)
+    context = apply_entry_filter_options(search, filter_options)
 
     query = (
         Query()
-        .WITH(("search", search.to_str(end='')))
+        .WITH(("search", str(search)))
         .SELECT(
             "search._id",
             "search._feed",
@@ -909,11 +988,51 @@ def make_search_entries_query(
             *"rank search._feed search._id".split(), keyword='HAVING'
         )
     elif sort == 'recent':
-        query.JOIN("entries ON (entries.id, entries.feed) = (_id, _feed)")
         apply_recent(query, keyword='HAVING', id_prefix='search._')
+    elif sort == 'random':
+        apply_random(query)
     else:
         assert False, "shouldn't get here"  # noqa: B011; # pragma: no cover
 
     log.debug("_search_entries query\n%s\n", query)
 
     return query, context
+
+
+def make_search_entry_counts_query(
+    filter_options: EntryFilterOptions,
+) -> Tuple[Query, Dict[str, Any]]:
+    query = (
+        Query()
+        .WITH(
+            (
+                "search",
+                """
+                SELECT _id, _feed
+                FROM entries_search
+                WHERE entries_search MATCH :query
+                GROUP BY _id, _feed
+                """,
+            )
+        )
+        .SELECT(
+            'count(*)',
+            'coalesce(sum(read == 1), 0)',
+            'coalesce(sum(important == 1), 0)',
+            """
+            coalesce(
+                sum(
+                    NOT (
+                        json_array_length(entries.enclosures) IS NULL OR json_array_length(entries.enclosures) = 0
+                    )
+                ), 0
+            )
+            """,
+        )
+        .FROM("entries")
+        .JOIN("search ON (id, feed) = (_id, _feed)")
+    )
+
+    context = apply_entry_filter_options(query, filter_options)
+
+    return query, context
