@@ -31,7 +31,9 @@ from ._types import FeedForUpdate
 from ._types import FeedUpdateIntent
 from ._types import TagFilter
 from ._utils import chunks
+from ._utils import exactly_one
 from ._utils import join_paginated_iter
+from ._utils import zero_or_one
 from .exceptions import EntryNotFoundError
 from .exceptions import FeedExistsError
 from .exceptions import FeedNotFoundError
@@ -40,9 +42,11 @@ from .exceptions import StorageError
 from .types import Content
 from .types import Enclosure
 from .types import Entry
+from .types import EntryCounts
 from .types import EntrySortOrder
 from .types import ExceptionInfo
 from .types import Feed
+from .types import FeedCounts
 from .types import FeedSortOrder
 from .types import JSONType
 
@@ -58,6 +62,7 @@ def create_db(db: sqlite3.Connection) -> None:
     create_entries(db)
     create_feed_metadata(db)
     create_feed_tags(db)
+    create_indexes(db)
 
 
 def create_feeds(db: sqlite3.Connection, name: str = 'feeds') -> None:
@@ -77,6 +82,7 @@ def create_feeds(db: sqlite3.Connection, name: str = 'feeds') -> None:
 
             -- reader data
             stale INTEGER NOT NULL DEFAULT 0,
+            updates_enabled INTEGER NOT NULL DEFAULT 1,
             last_updated TIMESTAMP,  -- null if the feed was never updated
             added TIMESTAMP NOT NULL,
             last_exception TEXT
@@ -156,6 +162,34 @@ def create_feed_tags(db: sqlite3.Connection) -> None:
     )
 
 
+def create_indexes(db: sqlite3.Connection) -> None:
+    # Speed up get_entries() queries that use apply_recent().
+    db.execute(
+        """
+        CREATE INDEX entries_by_kinda_first_updated ON entries(
+            first_updated_epoch,
+            coalesce(published, updated),
+            feed,
+            last_updated,
+            - feed_order,
+            id
+        );
+        """
+    )
+    db.execute(
+        """
+        CREATE INDEX entries_by_kinda_published ON entries (
+            coalesce(published, updated),
+            coalesce(published, updated),
+            feed,
+            last_updated,
+            - feed_order,
+            id
+        );
+        """
+    )
+
+
 def update_from_17_to_18(db: sqlite3.Connection) -> None:  # pragma: no cover
     # for https://github.com/lemon24/reader/issues/125
     db.execute("UPDATE feeds SET stale = 1;")
@@ -215,11 +249,21 @@ def update_from_23_to_24(db: sqlite3.Connection) -> None:  # pragma: no cover
     db.execute("ALTER TABLE entries ADD COLUMN original_feed TEXT;")
 
 
+def update_from_25_to_26(db: sqlite3.Connection) -> None:  # pragma: no cover
+    # for https://github.com/lemon24/reader/issues/187
+    db.execute(
+        """
+        ALTER TABLE feeds
+        ADD COLUMN updates_enabled INTEGER NOT NULL DEFAULT 1;
+        """
+    )
+
+
 def setup_db(db: sqlite3.Connection, wal_enabled: Optional[bool]) -> None:
     return setup_sqlite_db(
         db,
         create=create_db,
-        version=24,
+        version=26,
         migrations={
             # 1-9 removed before 0.1 (last in e4769d8ba77c61ec1fe2fbe99839e1826c17ace7)
             # 10-16 removed before 1.0 (last in 618f158ebc0034eefb724a55a84937d21c93c1a7)
@@ -233,6 +277,9 @@ def setup_db(db: sqlite3.Connection, wal_enabled: Optional[bool]) -> None:
             # for https://github.com/lemon24/reader/issues/149#issuecomment-700633577
             22: recreate_search_triggers,
             23: update_from_23_to_24,
+            # for https://github.com/lemon24/reader/issues/134#issuecomment-722454963
+            24: create_indexes,
+            25: update_from_25_to_26,
         },
         # Row value support was added in 3.15.
         # TODO: Remove the Search.update() check once this gets bumped to >=3.18.
@@ -261,6 +308,8 @@ def setup_db(db: sqlite3.Connection, wal_enabled: Optional[bool]) -> None:
 # have a look at the lessons here first:
 # https://github.com/lemon24/reader/issues/175#issuecomment-657495233
 
+# When adding a new method, add a new test_storage.py::test_errors_locked test.
+
 
 class Storage:
 
@@ -357,6 +406,8 @@ class Storage:
             else:
                 rowcount_exactly_one(cursor, lambda: FeedNotFoundError(old))
 
+            # Some of the fields are not kept from the old feed; details:
+            # https://github.com/lemon24/reader/issues/149#issuecomment-700532183
             self.db.execute(
                 """
                 UPDATE feeds
@@ -389,9 +440,33 @@ class Storage:
         self,
         filter_options: FeedFilterOptions = FeedFilterOptions(),  # noqa: B008
         sort: FeedSortOrder = 'title',
+        limit: Optional[int] = None,
+        starting_after: Optional[str] = None,
     ) -> Iterable[Feed]:
-        yield from join_paginated_iter(
-            partial(self.get_feeds_page, filter_options, sort), self.chunk_size,
+        rv = join_paginated_iter(
+            partial(self.get_feeds_page, filter_options, sort),  # type: ignore[arg-type]
+            self.chunk_size,
+            self.get_feed_last(sort, starting_after) if starting_after else None,
+            limit or 0,
+        )
+        yield from rv
+
+    @wrap_exceptions(StorageError)
+    def get_feed_last(self, sort: str, url: str) -> Tuple[Any, ...]:
+        # TODO: make this method private?
+
+        query = Query().FROM("feeds").WHERE("url = :url")
+
+        # TODO: kinda sorta duplicates the scrolling_window_order_by call
+        if sort == 'title':
+            query.SELECT(("kinda_title", "lower(coalesce(user_title, title))"), 'url')
+        elif sort == 'added':
+            query.SELECT("added", 'url')
+        else:
+            assert False, "shouldn't get here"  # noqa: B011; # pragma: no cover
+
+        return zero_or_one(
+            self.db.execute(str(query), dict(url=url)), lambda: FeedNotFoundError(url)
         )
 
     @wrap_exceptions_iter(StorageError)
@@ -402,8 +477,6 @@ class Storage:
         chunk_size: Optional[int] = None,
         last: Optional[_T] = None,
     ) -> Iterable[Tuple[Feed, Optional[_T]]]:
-        url, tags, broken = filter_options
-
         query = (
             Query()
             .SELECT(
@@ -416,19 +489,12 @@ class Storage:
                 'added',
                 'last_updated',
                 'last_exception',
+                'updates_enabled',
             )
             .FROM("feeds")
         )
-        context: Dict[str, object] = {}
 
-        if url:
-            query.WHERE("url = :url")
-            context.update(url=url)
-
-        context.update(apply_feed_tags_filter_options(query, tags, 'feeds.url'))
-
-        if broken is not None:
-            query.WHERE(f"last_exception IS {'NOT' if broken else ''}  NULL")
+        context = apply_feed_filter_options(query, filter_options)
 
         # sort by url at the end to make sure the order is deterministic
         if sort == 'title':
@@ -444,9 +510,33 @@ class Storage:
             self.db, query, context, feed_factory, chunk_size, last
         )
 
+    @wrap_exceptions(StorageError)
+    def get_feed_counts(
+        self,
+        filter_options: FeedFilterOptions = FeedFilterOptions(),  # noqa: B008
+    ) -> FeedCounts:
+        query = (
+            Query()
+            .SELECT(
+                'count(*)',
+                'coalesce(sum(last_exception IS NOT NULL), 0)',
+                'coalesce(sum(updates_enabled == 1), 0)',
+            )
+            .FROM("feeds")
+        )
+
+        context = apply_feed_filter_options(query, filter_options)
+
+        row = exactly_one(self.db.execute(str(query), context))
+
+        return FeedCounts(*row)
+
     @wrap_exceptions_iter(StorageError)
     def get_feeds_for_update(
-        self, url: Optional[str] = None, new_only: bool = False,
+        self,
+        url: Optional[str] = None,
+        new_only: bool = False,
+        enabled_only: bool = True,
     ) -> Iterable[FeedForUpdate]:
         # Reader shouldn't care this is paginated,
         # so we don't expose any pagination stuff.
@@ -476,6 +566,8 @@ class Storage:
                 context.update(url=url)
             if new_only:
                 query.WHERE("last_updated is NULL")
+            if enabled_only:
+                query.WHERE("updates_enabled")
 
             query.scrolling_window_order_by("url")
 
@@ -565,6 +657,15 @@ class Storage:
             )
         rowcount_exactly_one(cursor, lambda: FeedNotFoundError(url))
 
+    @wrap_exceptions(StorageError)
+    def set_feed_updates_enabled(self, url: str, enabled: bool) -> None:
+        with self.db:
+            cursor = self.db.execute(
+                "UPDATE feeds SET updates_enabled = :updates_enabled WHERE url = :url;",
+                dict(url=url, updates_enabled=enabled),
+            )
+            rowcount_exactly_one(cursor, lambda: FeedNotFoundError(url))
+
     @wrap_exceptions(StorageError)
     def mark_as_stale(self, url: str) -> None:
         with self.db:
@@ -818,20 +919,63 @@ class Storage:
         now: datetime,
         filter_options: EntryFilterOptions = EntryFilterOptions(),  # noqa: B008
         sort: EntrySortOrder = 'recent',
+        limit: Optional[int] = None,
+        starting_after: Optional[Tuple[str, str]] = None,
     ) -> Iterable[Entry]:
         # TODO: deduplicate
         if sort == 'recent':
-            yield from join_paginated_iter(
-                partial(self.get_entries_page, now, filter_options, sort),
+            rv = join_paginated_iter(
+                partial(self.get_entries_page, now, filter_options, sort),  # type: ignore[arg-type]
                 self.chunk_size,
+                self.get_entry_last(now, sort, starting_after)
+                if starting_after
+                else None,
+                limit or 0,
             )
         elif sort == 'random':
-            it = self.get_entries_page(now, filter_options, sort, self.chunk_size)
-            for entry, _ in it:
-                yield entry
+            assert not starting_after
+            it = self.get_entries_page(
+                now,
+                filter_options,
+                sort,
+                min(limit, self.chunk_size or limit) if limit else self.chunk_size,
+            )
+            rv = (entry for entry, _ in it)
         else:
             assert False, "shouldn't get here"  # noqa: B011; # pragma: no cover
 
+        yield from rv
+
+    @wrap_exceptions(StorageError)
+    def get_entry_last(
+        self, now: datetime, sort: str, entry: Tuple[str, str]
+    ) -> Tuple[Any, ...]:
+        # TODO: make this method private?
+
+        feed_url, entry_id = entry
+
+        query = Query().FROM("entries").WHERE("feed = :feed AND id = :id")
+
+        assert sort != 'random'
+
+        # TODO: kinda sorta duplicates the scrolling_window_order_by call
+        if sort == 'recent':
+            query.SELECT(*make_recent_last_select())
+        else:
+            assert False, "shouldn't get here"  # noqa: B011; # pragma: no cover
+
+        context = dict(
+            feed=feed_url,
+            id=entry_id,
+            # TODO: duplicated from get_entries_page()
+            recent_threshold=now - self.recent_threshold,
+        )
+
+        return zero_or_one(
+            self.db.execute(str(query), context),
+            lambda: EntryNotFoundError(feed_url, entry_id),
+        )
+
     @wrap_exceptions_iter(StorageError)
     def get_entries_page(
         self,
@@ -848,19 +992,18 @@ class Storage:
 
         context = dict(
             recent_threshold=now - self.recent_threshold,
-            **filter_options._asdict(),
             **query_context,
         )
 
         def value_factory(t: Tuple[Any, ...]) -> Entry:
-            feed = feed_factory(t[0:9])
-            entry = t[9:16] + (
-                tuple(Content(**d) for d in json.loads(t[16])) if t[16] else (),
-                tuple(Enclosure(**d) for d in json.loads(t[17])) if t[17] else (),
-                t[18] == 1,
+            feed = feed_factory(t[0:10])
+            entry = t[10:17] + (
+                tuple(Content(**d) for d in json.loads(t[17])) if t[17] else (),
+                tuple(Enclosure(**d) for d in json.loads(t[18])) if t[18] else (),
                 t[19] == 1,
-                t[20],
-                t[21] or feed.url,
+                t[20] == 1,
+                t[21],
+                t[22] or feed.url,
                 feed,
             )
             return Entry._make(entry)
@@ -869,11 +1012,44 @@ class Storage:
             self.db, query, context, value_factory, chunk_size, last
         )
 
+    @wrap_exceptions(StorageError)
+    def get_entry_counts(
+        self,
+        filter_options: EntryFilterOptions = EntryFilterOptions(),  # noqa: B008
+    ) -> EntryCounts:
+        query = (
+            Query()
+            .SELECT(
+                'count(*)',
+                'coalesce(sum(read == 1), 0)',
+                'coalesce(sum(important == 1), 0)',
+                """
+                coalesce(
+                    sum(
+                        NOT (
+                            json_array_length(entries.enclosures) IS NULL OR json_array_length(entries.enclosures) = 0
+                        )
+                    ), 0
+                )
+                """,
+            )
+            .FROM("entries")
+        )
+
+        context = apply_entry_filter_options(query, filter_options)
+
+        row = exactly_one(self.db.execute(str(query), context))
+
+        return EntryCounts(*row)
+
     def iter_feed_metadata(
-        self, feed_url: str, key: Optional[str] = None,
+        self,
+        feed_url: str,
+        key: Optional[str] = None,
     ) -> Iterable[Tuple[str, JSONType]]:
         yield from join_paginated_iter(
-            partial(self.iter_feed_metadata_page, feed_url, key), self.chunk_size,
+            partial(self.iter_feed_metadata_page, feed_url, key),
+            self.chunk_size,
         )
 
     @wrap_exceptions_iter(StorageError)
@@ -957,12 +1133,14 @@ class Storage:
     def remove_feed_tag(self, feed_url: str, tag: str) -> None:
         with self.db:
             self.db.execute(
-                "DELETE FROM feed_tags WHERE feed = ? AND tag = ?;", (feed_url, tag),
+                "DELETE FROM feed_tags WHERE feed = ? AND tag = ?;",
+                (feed_url, tag),
             )
 
     def get_feed_tags(self, feed_url: Optional[str]) -> Iterable[str]:
         yield from join_paginated_iter(
-            partial(self.get_feed_tags_page, feed_url), self.chunk_size,
+            partial(self.get_feed_tags_page, feed_url),
+            self.chunk_size,
         )
 
     @wrap_exceptions_iter(StorageError)
@@ -992,11 +1170,40 @@ class Storage:
 
 
 def feed_factory(t: Tuple[Any, ...]) -> Feed:
-    return Feed._make(t[:8] + (ExceptionInfo(**json.loads(t[8])) if t[8] else None,))
+    return Feed._make(
+        t[:8]
+        + (
+            ExceptionInfo(**json.loads(t[8])) if t[8] else None,
+            t[9] == 1,
+        )
+    )
+
+
+def apply_feed_filter_options(
+    query: Query,
+    filter_options: FeedFilterOptions,
+) -> Dict[str, Any]:
+    url, tags, broken, updates_enabled = filter_options
+
+    context: Dict[str, object] = {}
+
+    if url:
+        query.WHERE("url = :url")
+        context.update(url=url)
+
+    context.update(apply_feed_tags_filter_options(query, tags, 'feeds.url'))
+
+    if broken is not None:
+        query.WHERE(f"last_exception IS {'NOT' if broken else ''} NULL")
+    if updates_enabled is not None:
+        query.WHERE(f"{'' if updates_enabled else 'NOT'} updates_enabled")
+
+    return context
 
 
 def make_get_entries_query(
-    filter_options: EntryFilterOptions, sort: EntrySortOrder,
+    filter_options: EntryFilterOptions,
+    sort: EntrySortOrder,
 ) -> Tuple[Query, Dict[str, Any]]:
     query = (
         Query()
@@ -1011,6 +1218,7 @@ def make_get_entries_query(
             feeds.added
             feeds.last_updated
             feeds.last_exception
+            feeds.updates_enabled
             entries.id
             entries.updated
             entries.title
@@ -1030,15 +1238,13 @@ def make_get_entries_query(
         .JOIN("feeds ON feeds.url = entries.feed")
     )
 
-    filter_context = apply_filter_options(query, filter_options)
+    filter_context = apply_entry_filter_options(query, filter_options)
 
     if sort == 'recent':
         apply_recent(query)
 
     elif sort == 'random':
-        # TODO: "order by random()" always goes through the full result set, which is inefficient
-        # details here https://github.com/lemon24/reader/issues/105#issue-409493128
-        query.ORDER_BY("random()")
+        apply_random(query)
 
     else:
         assert False, "shouldn't get here"  # noqa: B011; # pragma: no cover
@@ -1048,16 +1254,20 @@ def make_get_entries_query(
     return query, filter_context
 
 
-def apply_filter_options(
+def apply_entry_filter_options(
     query: Query, filter_options: EntryFilterOptions, keyword: str = 'WHERE'
 ) -> Dict[str, Any]:
     add = getattr(query, keyword)
     feed_url, entry_id, read, important, has_enclosures, feed_tags = filter_options
 
+    context = {}
+
     if feed_url:
         add("entries.feed = :feed_url")
+        context['feed_url'] = feed_url
         if entry_id:
             add("entries.id = :entry_id")
+            context['entry_id'] = entry_id
 
     if read is not None:
         add(f"{'' if read else 'NOT'} entries.read")
@@ -1074,15 +1284,20 @@ def apply_filter_options(
             """
         )
 
-    context = apply_feed_tags_filter_options(
-        query, feed_tags, 'entries.feed', keyword=keyword
+    context.update(
+        apply_feed_tags_filter_options(
+            query, feed_tags, 'entries.feed', keyword=keyword
+        )
     )
 
     return context
 
 
 def apply_feed_tags_filter_options(
-    query: Query, tags: TagFilter, url_column: str, keyword: str = 'WHERE',
+    query: Query,
+    tags: TagFilter,
+    url_column: str,
+    keyword: str = 'WHERE',
 ) -> Dict[str, Any]:
     add = getattr(query, keyword)
 
@@ -1127,13 +1342,10 @@ def apply_feed_tags_filter_options(
     return context
 
 
-def apply_recent(
-    query: Query, keyword: str = 'WHERE', id_prefix: str = 'entries.'
-) -> None:
-    query.SELECT(
-        "entries.last_updated",
+def make_recent_last_select(id_prefix: str = 'entries.') -> Sequence[Any]:
+    return [
         (
-            "kinda_first_updated",
+            'kinda_first_updated',
             """
             coalesce (
                 CASE
@@ -1146,18 +1358,93 @@ def apply_recent(
             )
             """,
         ),
-        ("kinda_published", "coalesce(entries.published, entries.updated)"),
-        ("negative_feed_order", "- entries.feed_order"),
+        ('kinda_published', 'coalesce(published, updated)'),
+        f'{id_prefix}feed',
+        'last_updated',
+        ('negative_feed_order', '- feed_order'),
+        f'{id_prefix}id',
+    ]
+
+
+def apply_recent(
+    query: Query, keyword: str = 'WHERE', id_prefix: str = 'entries.'
+) -> None:
+    """Change query to sort entries by "recent"."""
+
+    # Version of apply_recent() that takes advantage of indexes, implemented in
+    # <https://github.com/lemon24/reader/issues/134#issuecomment-722454963>.
+    # An older version that does not can be found in reader 1.9.
+    #
+    # WARNING: Always keep the entries_by_kinda_* indexes in sync with the ORDER BY of the by_kinda_* CTEs below.
+
+    def make_by_kinda_cte(recent: bool) -> str:
+        if recent:
+            kinda_first_updated = 'first_updated_epoch'
+            sign = '>='
+        else:
+            kinda_first_updated = 'coalesce(published, updated)'
+            sign = '<'
+
+        return f"""
+            SELECT
+                feed,
+                id,
+                last_updated,
+                {kinda_first_updated} as kinda_first_updated,
+                coalesce(published, updated) as kinda_published,
+                - feed_order as negative_feed_order
+            FROM entries
+            WHERE kinda_published {sign} :recent_threshold
+            ORDER BY
+                kinda_first_updated DESC,
+                kinda_published DESC,
+                feed DESC,
+                last_updated DESC,
+                negative_feed_order DESC,
+                id DESC
+        """
+
+    query.WITH(
+        ('by_kinda_first_updated', make_by_kinda_cte(recent=True)),
+        ('by_kinda_published', make_by_kinda_cte(recent=False)),
+        (
+            'ids',
+            """
+            SELECT * FROM by_kinda_first_updated
+            UNION ALL
+            SELECT * FROM by_kinda_published
+            """,
+        ),
     )
+    query.JOIN(f"ids ON (ids.id, ids.feed) = ({id_prefix}id, {id_prefix}feed)")
+
+    query.SELECT(
+        "ids.last_updated",
+        "kinda_first_updated",
+        "kinda_published",
+        "negative_feed_order",
+    )
+
     query.scrolling_window_order_by(
         *f"""
         kinda_first_updated
         kinda_published
         {id_prefix}feed
-        entries.last_updated
+        ids.last_updated
         negative_feed_order
         {id_prefix}id
         """.split(),
         desc=True,
         keyword=keyword,
     )
+
+
+def apply_random(query: Query) -> None:
+    # TODO: "order by random()" always goes through the full result set,
+    # which is inefficient; details:
+    # https://github.com/lemon24/reader/issues/105#issue-409493128
+    #
+    # This is a separate function in the hope that search
+    # can benefit from future optimizations.
+    #
+    query.ORDER_BY("random()")
