@@ -2,6 +2,7 @@ import builtins
 import datetime
 import itertools
 import logging
+import numbers
 import warnings
 from typing import Any
 from typing import Callable
@@ -29,6 +30,7 @@ from ._types import ParsedFeed
 from ._utils import make_noop_context_manager
 from ._utils import make_pool_map
 from ._utils import zero_or_one
+from .exceptions import _NotModified
 from .exceptions import EntryNotFoundError
 from .exceptions import FeedNotFoundError
 from .exceptions import MetadataNotFoundError
@@ -36,10 +38,13 @@ from .exceptions import ParseError
 from .types import _entry_argument
 from .types import _feed_argument
 from .types import Entry
+from .types import EntryCounts
 from .types import EntryInput
+from .types import EntrySearchCounts
 from .types import EntrySearchResult
 from .types import EntrySortOrder
 from .types import Feed
+from .types import FeedCounts
 from .types import FeedInput
 from .types import FeedSortOrder
 from .types import JSONType
@@ -164,11 +169,33 @@ class Reader:
 
     """A feed reader.
 
+    Persists feed and entry state, provides operations on them,
+    and stores configuration.
+
+    Currently, the following feed types are supported:
+
+    * Atom (provided by `feedparser`_)
+    * RSS (provided by `feedparser`_)
+    * JSON Feed
+
+    .. _feedparser: https://feedparser.readthedocs.io/en/latest/
+
+
     .. important::
 
         Reader objects should be created using :func:`make_reader`; the Reader
         constructor is not stable yet and may change without any notice.
 
+    .. important::
+
+        The :class:`Reader` object is not thread safe;
+        its methods should be called only from the thread that created it.
+
+        To access the same database from multiple threads,
+        create one instance in each thread.
+        If you have a strong use case preventing you to do so,
+        please +1 / comment in :issue:`206`.
+
     """
 
     def __init__(
@@ -207,6 +234,8 @@ class Reader:
     def add_feed(self, feed: FeedInput) -> None:
         """Add a new feed.
 
+        Feed updates are enabled by default.
+
         Args:
             feed (str or Feed): The feed URL.
 
@@ -268,7 +297,10 @@ class Reader:
         feed: Optional[FeedInput] = None,
         tags: TagFilterInput = None,
         broken: Optional[bool] = None,
+        updates_enabled: Optional[bool] = None,
         sort: FeedSortOrder = 'title',
+        limit: Optional[int] = None,
+        starting_after: Optional[FeedInput] = None,
     ) -> Iterable[Feed]:
         """Get all or some of the feeds.
 
@@ -325,15 +357,22 @@ class Reader:
             tags (None or bool or list(str or bool or list(str or bool))):
                 Only return feeds matching these tags.
             broken (bool or None): Only return broken / healthy feeds.
+            updates_enabled (bool or None):
+                Only return feeds that have updates enabled / disabled.
             sort (str): How to order feeds; one of ``'title'`` (by
                 :attr:`~Feed.user_title` or :attr:`~Feed.title`, case
                 insensitive; default), or ``'added'`` (last added first).
+            limit (int or None): A limit on the number of feeds to be returned;
+                by default, all feeds are returned.
+            starting_after (str or Feed or None):
+                Return feeds after this feed; a cursor for use in pagination.
 
         Yields:
             :class:`Feed`: Sorted according to ``sort``.
 
         Raises:
             StorageError
+            FeedNotFoundError: If ``starting_after`` does not exist.
 
         .. versionadded:: 1.7
             The ``tags`` keyword argument.
@@ -341,13 +380,30 @@ class Reader:
         .. versionadded:: 1.7
             The ``broken`` keyword argument.
 
+        .. versionadded:: 1.11
+            The ``updates_enabled`` keyword argument.
+
+        .. versionadded:: 1.12
+            The ``limit`` and ``starting_after`` keyword arguments.
+
         """
-        filter_options = FeedFilterOptions.from_args(feed, tags, broken)
+        filter_options = FeedFilterOptions.from_args(
+            feed, tags, broken, updates_enabled
+        )
 
         if sort not in ('title', 'added'):
             raise ValueError("sort should be one of ('title', 'added')")
 
-        return self._storage.get_feeds(filter_options, sort)
+        if limit is not None:
+            if not isinstance(limit, numbers.Integral) or limit < 1:
+                raise ValueError("limit should be a positive integer")
+
+        return self._storage.get_feeds(
+            filter_options,
+            sort,
+            limit,
+            _feed_argument(starting_after) if starting_after else None,
+        )
 
     @overload
     def get_feed(self, feed: FeedInput) -> Feed:  # pragma: no cover
@@ -385,6 +441,40 @@ class Reader:
             default,
         )
 
+    def get_feed_counts(
+        self,
+        *,
+        feed: Optional[FeedInput] = None,
+        tags: TagFilterInput = None,
+        broken: Optional[bool] = None,
+        updates_enabled: Optional[bool] = None,
+    ) -> FeedCounts:
+        """Count all or some of the feeds.
+
+        See :meth:`~Reader.get_feeds()` for details on how filtering works.
+
+        Args:
+            feed (str or Feed or None): Only count the feed with this URL.
+            tags (None or bool or list(str or bool or list(str or bool))):
+                Only count feeds matching these tags.
+            broken (bool or None): Only count broken / healthy feeds.
+            updates_enabled (bool or None):
+                Only count feeds that have updates enabled / disabled.
+
+        Returns:
+            :class:`FeedCounts`:
+
+        Raises:
+            StorageError
+
+        .. versionadded:: 1.11
+
+        """
+        filter_options = FeedFilterOptions.from_args(
+            feed, tags, broken, updates_enabled
+        )
+        return self._storage.get_feed_counts(filter_options)
+
     def set_feed_user_title(self, feed: FeedInput, title: Optional[str]) -> None:
         """Set a user-defined title for a feed.
 
@@ -400,8 +490,46 @@ class Reader:
         url = _feed_argument(feed)
         return self._storage.set_feed_user_title(url, title)
 
+    def enable_feed_updates(self, feed: FeedInput) -> None:
+        """Enable updates for a feed.
+
+        See :meth:`~Reader.update_feeds` for details.
+
+        Args:
+            feed (str or Feed): The feed URL.
+
+        Raises:
+            FeedNotFoundError
+            StorageError
+
+        .. versionadded:: 1.11
+
+        """
+        url = _feed_argument(feed)
+        self._storage.set_feed_updates_enabled(url, True)
+
+    def disable_feed_updates(self, feed: FeedInput) -> None:
+        """Disable updates for a feed.
+
+        See :meth:`~Reader.update_feeds` for details.
+
+        Args:
+            feed (str or Feed): The feed URL.
+
+        Raises:
+            FeedNotFoundError
+            StorageError
+
+        .. versionadded:: 1.11
+
+        """
+        url = _feed_argument(feed)
+        self._storage.set_feed_updates_enabled(url, False)
+
     def update_feeds(self, new_only: bool = False, workers: int = 1) -> None:
-        """Update all the feeds.
+        """Update all the feeds that have updates enabled.
+
+        Silently skip feeds that raise :exc:`ParseError`.
 
         Args:
             new_only (bool): Only update feeds that have never been updated.
@@ -410,6 +538,9 @@ class Reader:
         Raises:
             StorageError
 
+        .. versionchanged:: 1.11
+            Only update the feeds that have updates enabled.
+
         """
 
         if workers < 1:
@@ -443,6 +574,8 @@ class Reader:
     def update_feed(self, feed: FeedInput) -> None:
         """Update a single feed.
 
+        The feed will be updated even if updates are disabled for it.
+
         Args:
             feed (str or Feed): The feed URL.
 
@@ -453,7 +586,10 @@ class Reader:
 
         """
         url = _feed_argument(feed)
-        exc = zero_or_one(self._update_feeds(url=url), lambda: FeedNotFoundError(url),)
+        exc = zero_or_one(
+            self._update_feeds(url=url, enabled_only=False),
+            lambda: FeedNotFoundError(url),
+        )
         if exc:
             raise exc
 
@@ -474,6 +610,7 @@ class Reader:
         self,
         url: Optional[str] = None,
         new_only: bool = False,
+        enabled_only: bool = True,
         map: Callable[[Callable[[Any], Any], Iterable[Any]], Iterator[Any]] = map,
     ) -> Iterator[Optional[Exception]]:
 
@@ -505,7 +642,9 @@ class Reader:
         # It may make sense to also have _get_entries_for_update run in
         # parallel with a different (slower) storage, but for now we're good.
 
-        feeds_for_update = self._storage.get_feeds_for_update(url, new_only)
+        feeds_for_update = self._storage.get_feeds_for_update(
+            url, new_only, enabled_only
+        )
         feeds_for_update = builtins.map(
             self._updater.process_old_feed, feeds_for_update
         )
@@ -544,10 +683,10 @@ class Reader:
 
     def _parse_feed_for_update(
         self, feed: FeedForUpdate
-    ) -> Tuple[FeedForUpdate, Union[ParsedFeed, Exception]]:
+    ) -> Tuple[FeedForUpdate, Union[ParsedFeed, ParseError, _NotModified]]:
         try:
             return feed, self._parser(feed.url, feed.http_etag, feed.http_last_modified)
-        except Exception as e:
+        except (ParseError, _NotModified) as e:
             log.debug(
                 "_parse_feed_for_update exception, traceback follows", exc_info=True
             )
@@ -582,6 +721,8 @@ class Reader:
         has_enclosures: Optional[bool] = None,
         feed_tags: TagFilterInput = None,
         sort: EntrySortOrder = 'recent',
+        limit: Optional[int] = None,
+        starting_after: Optional[EntryInput] = None,
     ) -> Iterable[Entry]:
         """Get all or some of the entries.
 
@@ -605,7 +746,7 @@ class Reader:
 
         ``'random'``
 
-            Random. At at most 256 entries will be returned.
+            Random order (shuffled). At at most 256 entries will be returned.
 
             .. versionadded:: 1.2
 
@@ -622,12 +763,18 @@ class Reader:
                 works like the :meth:`~Reader.get_feeds()` ``tags`` argument.
             sort (str): How to order entries; one of ``'recent'`` (default)
                 or ``'random'``.
+            limit (int or None): A limit on the number of entries to be returned;
+                by default, all entries are returned.
+            starting_after (tuple(str, str) or Entry or None):
+                Return entries after this entry; a cursor for use in pagination.
+                Using ``starting_after`` with ``sort='random'`` is not supported.
 
         Yields:
             :class:`Entry`: Sorted according to ``sort``.
 
         Raises:
             StorageError
+            EntryNotFoundError: If ``starting_after`` does not exist.
 
         .. versionadded:: 1.2
             The ``sort`` keyword argument.
@@ -635,6 +782,9 @@ class Reader:
         .. versionadded:: 1.7
             The ``feed_tags`` keyword argument.
 
+        .. versionadded:: 1.12
+            The ``limit`` and ``starting_after`` keyword arguments.
+
         """
 
         # If we ever implement pagination, consider following the guidance in
@@ -643,10 +793,25 @@ class Reader:
         filter_options = EntryFilterOptions.from_args(
             feed, entry, read, important, has_enclosures, feed_tags
         )
+
         if sort not in ('recent', 'random'):
             raise ValueError("sort should be one of ('recent', 'random')")
+
+        if limit is not None:
+            if not isinstance(limit, numbers.Integral) or limit < 1:
+                raise ValueError("limit should be a positive integer")
+
+        if starting_after and sort == 'random':
+            raise ValueError("using starting_after with sort='random' not supported")
+
         now = self._now()
-        return self._storage.get_entries(now, filter_options, sort)
+        return self._storage.get_entries(
+            now,
+            filter_options,
+            sort,
+            limit,
+            _entry_argument(starting_after) if starting_after else None,
+        )
 
     @overload
     def get_entry(self, entry: EntryInput) -> Entry:  # pragma: no cover
@@ -684,6 +849,46 @@ class Reader:
             default,
         )
 
+    def get_entry_counts(
+        self,
+        *,
+        feed: Optional[FeedInput] = None,
+        entry: Optional[EntryInput] = None,
+        read: Optional[bool] = None,
+        important: Optional[bool] = None,
+        has_enclosures: Optional[bool] = None,
+        feed_tags: TagFilterInput = None,
+    ) -> EntryCounts:
+        """Count all or some of the entries.
+
+        See :meth:`~Reader.get_entries()` for details on how filtering works.
+
+        Args:
+            feed (str or Feed or None): Only count the entries for this feed.
+            entry (tuple(str, str) or Entry or None):
+                Only count the entry with this (feed URL, entry id) tuple.
+            read (bool or None): Only count (un)read entries.
+            important (bool or None): Only count (un)important entries.
+            has_enclosures (bool or None): Only count entries that (don't)
+                have enclosures.
+            feed_tags (None or bool or list(str or bool or list(str or bool))):
+                Only count the entries from feeds matching these tags.
+
+        Returns:
+            :class:`EntryCounts`:
+
+        Raises:
+            StorageError
+
+        .. versionadded:: 1.11
+
+        """
+
+        filter_options = EntryFilterOptions.from_args(
+            feed, entry, read, important, has_enclosures, feed_tags
+        )
+        return self._storage.get_entry_counts(filter_options)
+
     def mark_as_read(self, entry: EntryInput) -> None:
         """Mark an entry as read.
 
@@ -741,7 +946,10 @@ class Reader:
         self._storage.mark_as_important_unimportant(feed_url, entry_id, False)
 
     def iter_feed_metadata(
-        self, feed: FeedInput, *, key: Optional[str] = None,
+        self,
+        feed: FeedInput,
+        *,
+        key: Optional[str] = None,
     ) -> Iterable[Tuple[str, JSONType]]:
         """Get all or some of the metadata values for a feed.
 
@@ -891,6 +1099,8 @@ class Reader:
         has_enclosures: Optional[bool] = None,
         feed_tags: TagFilterInput = None,
         sort: SearchSortOrder = 'relevant',
+        limit: Optional[int] = None,
+        starting_after: Optional[EntryInput] = None,
     ) -> Iterable[EntrySearchResult]:
         """Get entries matching a full-text search query.
 
@@ -907,6 +1117,12 @@ class Reader:
 
             .. versionadded:: 1.4
 
+        ``'random'``
+
+            Random order (shuffled). At at most 256 entries will be returned.
+
+            .. versionadded:: 1.10
+
         Note:
             The query syntax is dependent on the search provider.
 
@@ -946,8 +1162,13 @@ class Reader:
             feed_tags (None or bool or list(str or bool or list(str or bool))):
                 Only return the entries from feeds matching these tags;
                 works like the :meth:`~Reader.get_feeds()` ``tags`` argument.
-            sort (str): How to order results; one of ``'relevant'`` (default)
-                or ``'recent'``.
+            sort (str): How to order results; one of ``'relevant'`` (default),
+                ``'recent'``, or ``'random'``.
+            limit (int or None): A limit on the number of results to be returned;
+                by default, all results are returned.
+            starting_after (tuple(str, str) or EntrySearchResult or None):
+                Return results after this result; a cursor for use in pagination.
+                Using ``starting_after`` with ``sort='random'`` is not supported.
 
         Yields:
             :class:`EntrySearchResult`: Sorted according to ``sort``.
@@ -957,6 +1178,7 @@ class Reader:
             InvalidSearchQueryError
             SearchError
             StorageError
+            EntryNotFoundError: If ``starting_after`` does not exist.
 
         .. versionadded:: 1.4
             The ``sort`` keyword argument.
@@ -964,14 +1186,81 @@ class Reader:
         .. versionadded:: 1.7
             The ``feed_tags`` keyword argument.
 
+        .. versionadded:: 1.12
+            The ``limit`` and ``starting_after`` keyword arguments.
+
         """
         filter_options = EntryFilterOptions.from_args(
             feed, entry, read, important, has_enclosures, feed_tags
         )
-        if sort not in ('relevant', 'recent'):
-            raise ValueError("sort should be one of ('relevant', 'recent')")
+
+        if sort not in ('relevant', 'recent', 'random'):
+            raise ValueError("sort should be one of ('relevant', 'recent', 'random')")
+
+        if limit is not None:
+            if not isinstance(limit, numbers.Integral) or limit < 1:
+                raise ValueError("limit should be a positive integer")
+
+        if starting_after and sort == 'random':
+            raise ValueError("using starting_after with sort='random' not supported")
+
         now = self._now()
-        return self._search.search_entries(query, now, filter_options, sort)
+        return self._search.search_entries(
+            query,
+            now,
+            filter_options,
+            sort,
+            limit,
+            _entry_argument(starting_after) if starting_after else None,
+        )
+
+    def search_entry_counts(
+        self,
+        query: str,
+        *,
+        feed: Optional[FeedInput] = None,
+        entry: Optional[EntryInput] = None,
+        read: Optional[bool] = None,
+        important: Optional[bool] = None,
+        has_enclosures: Optional[bool] = None,
+        feed_tags: TagFilterInput = None,
+    ) -> EntrySearchCounts:
+        """Count entries matching a full-text search query.
+
+        See :meth:`~Reader.search_entries()` for details on how
+        the query syntax and filtering work.
+
+        Search must be enabled to call this method.
+
+        Args:
+            query (str): The search query.
+            feed (str or Feed or None): Only count the entries for this feed.
+            entry (tuple(str, str) or Entry or None):
+                Only count the entry with this (feed URL, entry id) tuple.
+            read (bool or None): Only count (un)read entries.
+            important (bool or None): Only count (un)important entries.
+            has_enclosures (bool or None): Only count entries that (don't)
+                have enclosures.
+            feed_tags (None or bool or list(str or bool or list(str or bool))):
+                Only count the entries from feeds matching these tags.
+
+        Returns:
+            :class:`EntrySearchCounts`:
+
+        Raises:
+            SearchNotEnabledError
+            InvalidSearchQueryError
+            SearchError
+            StorageError
+
+        .. versionadded:: 1.11
+
+        """
+
+        filter_options = EntryFilterOptions.from_args(
+            feed, entry, read, important, has_enclosures, feed_tags
+        )
+        return self._search.search_entry_counts(query, filter_options)
 
     def add_feed_tag(self, feed: FeedInput, tag: str) -> None:
         """Add a tag to a feed.
