@@ -1,5 +1,7 @@
 import io
+import json
 import logging
+from contextlib import contextmanager
 from unittest.mock import MagicMock
 
 import feedparser
@@ -9,9 +11,13 @@ from utils import make_url_base
 
 from reader import Feed
 from reader._parser import default_parser
-from reader._parser import FileParser
-from reader._parser import parse_feed
+from reader._parser import FeedparserParser
+from reader._parser import FileRetriever
+from reader._parser import JSONFeedParser
+from reader._parser import Parser
+from reader._parser import RetrieveResult
 from reader._parser import SessionWrapper
+from reader._types import FeedData
 from reader.exceptions import _NotModified
 from reader.exceptions import ParseError
 
@@ -22,6 +28,21 @@ def parse():
     yield parse
 
 
+def str_bytes_parser(parse):
+    def wrapper(url, file, headers=None):
+        if isinstance(file, str):
+            file = io.BytesIO(file.encode('utf-8'))
+        elif isinstance(file, bytes):
+            file = io.BytesIO(file)
+        return parse(url, file, headers)
+
+    return wrapper
+
+
+feedparser_parse = str_bytes_parser(FeedparserParser())
+jsonfeed_parse = str_bytes_parser(JSONFeedParser())
+
+
 def _make_relative_path_url(**_):
     def make_url(feed_path):
         return feed_path.relto(feed_path.join('../..'))
@@ -44,10 +65,14 @@ def _make_http_url(requests_mock, **_):
         url = 'http://example.com/' + feed_path.basename
         headers = {}
         if feed_path.ext == '.rss':
-            headers['Content-Type'] = 'application/x-rss+xml'
+            headers['Content-Type'] = 'application/rss+xml'
         elif feed_path.ext == '.atom':
             headers['Content-Type'] = 'application/atom+xml'
-        requests_mock.get(url, text=feed_path.read(), headers=headers)
+        elif feed_path.ext == '.json':
+            headers['Content-Type'] = 'application/feed+json'
+        with open(str(feed_path), 'rb') as f:
+            body = f.read()
+        requests_mock.get(url, content=body, headers=headers)
         return url
 
     return make_url
@@ -61,10 +86,14 @@ def _make_https_url(requests_mock, **_):
         url = 'https://example.com/' + feed_path.basename
         headers = {}
         if feed_path.ext == '.rss':
-            headers['Content-Type'] = 'application/x-rss+xml'
+            headers['Content-Type'] = 'application/rss+xml'
         elif feed_path.ext == '.atom':
             headers['Content-Type'] = 'application/atom+xml'
-        requests_mock.get(url, text=feed_path.read(), headers=headers)
+        elif feed_path.ext == '.json':
+            headers['Content-Type'] = 'application/feed+json'
+        with open(str(feed_path), 'rb') as f:
+            body = f.read()
+        requests_mock.get(url, content=body, headers=headers)
         return url
 
     return make_url
@@ -75,16 +104,20 @@ def _make_http_gzip_url(requests_mock, **_):
         url = 'http://example.com/' + feed_path.basename
         headers = {}
         if feed_path.ext == '.rss':
-            headers['Content-Type'] = 'application/x-rss+xml'
+            headers['Content-Type'] = 'application/rss+xml'
         elif feed_path.ext == '.atom':
             headers['Content-Type'] = 'application/atom+xml'
+        elif feed_path.ext == '.json':
+            headers['Content-Type'] = 'application/feed+json'
         headers['Content-Encoding'] = 'gzip'
+        with open(str(feed_path), 'rb') as f:
+            body = f.read()
 
         import io, gzip
 
         compressed_file = io.BytesIO()
         gz = gzip.GzipFile(fileobj=compressed_file, mode='wb')
-        gz.write(feed_path.read_binary())
+        gz.write(body)
         gz.close()
 
         requests_mock.get(url, content=compressed_file.getvalue(), headers=headers)
@@ -96,7 +129,9 @@ def _make_http_gzip_url(requests_mock, **_):
 def _make_http_url_missing_content_type(requests_mock, **_):
     def make_url(feed_path):
         url = 'http://example.com/' + feed_path.basename
-        requests_mock.get(url, text=feed_path.read())
+        with open(str(feed_path), 'rb') as f:
+            body = f.read()
+        requests_mock.get(url, content=body)
         return url
 
     return make_url
@@ -116,8 +151,11 @@ def make_url(request, requests_mock):
     return request.param(requests_mock=requests_mock)
 
 
-@pytest.mark.parametrize('feed_type', ['rss', 'atom'])
-@pytest.mark.parametrize('data_file', ['full', 'empty', 'relative'])
+@pytest.mark.parametrize(
+    'feed_type, data_file',
+    [(ft, df) for ft in ['rss', 'atom'] for df in ['full', 'empty', 'relative']]
+    + [('json', df) for df in ['full', 'empty', 'invalid']],
+)
 def test_parse(monkeypatch, feed_type, data_file, parse, make_url, data_dir):
     monkeypatch.chdir(data_dir.dirname)
 
@@ -141,6 +179,28 @@ def test_parse(monkeypatch, feed_type, data_file, parse, make_url, data_dir):
     assert entries == expected['entries']
 
 
+def test_no_mime_type(monkeypatch, parse, make_url, data_dir):
+    """Like test_parse with _make_http_url_missing_content_type,
+    but with an URL parser.
+    """
+    monkeypatch.chdir(data_dir.dirname)
+    feed_path = data_dir.join('custom')
+    feed_url = make_url(feed_path)
+
+    def custom_parser(url, file, headers=None):
+        return FeedData(url=url, title=file.read().decode('utf-8')), []
+
+    parse.mount_parser_by_url(feed_url, custom_parser)
+
+    feed, entries, _, _ = parse(feed_url)
+
+    with open(str(feed_path), encoding='utf-8') as f:
+        expected_feed = FeedData(url=feed_url, title=f.read())
+
+    assert feed == expected_feed
+    assert entries == []
+
+
 def test_feedparser_exceptions(monkeypatch, parse, data_dir):
     """parse() should reraise most feedparser exceptions."""
 
@@ -235,7 +295,7 @@ def make_http_get_headers_url(requests_mock):
         url = 'http://example.com/' + feed_path.basename
         headers = {}
         if feed_path.ext == '.rss':
-            headers['Content-Type'] = 'application/x-rss+xml'
+            headers['Content-Type'] = 'application/rss+xml'
         elif feed_path.ext == '.atom':
             headers['Content-Type'] = 'application/atom+xml'
 
@@ -249,9 +309,12 @@ def make_http_get_headers_url(requests_mock):
     yield make_url
 
 
-@pytest.mark.parametrize('feed_type', ['rss', 'atom'])
+@pytest.mark.parametrize('feed_type', ['rss', 'atom', 'json'])
 def test_parse_sends_etag_last_modified(
-    parse, make_http_get_headers_url, data_dir, feed_type,
+    parse,
+    make_http_get_headers_url,
+    data_dir,
+    feed_type,
 ):
     feed_url = make_http_get_headers_url(data_dir.join('full.' + feed_type))
     parse(feed_url, 'etag', 'last_modified')
@@ -268,7 +331,7 @@ def make_http_etag_last_modified_url(requests_mock):
         url = 'http://example.com/' + feed_path.basename
         headers = {'ETag': make_url.etag, 'Last-Modified': make_url.last_modified}
         if feed_path.ext == '.rss':
-            headers['Content-Type'] = 'application/x-rss+xml'
+            headers['Content-Type'] = 'application/rss+xml'
         elif feed_path.ext == '.atom':
             headers['Content-Type'] = 'application/atom+xml'
         requests_mock.get(url, text=feed_path.read(), headers=headers)
@@ -277,7 +340,7 @@ def make_http_etag_last_modified_url(requests_mock):
     yield make_url
 
 
-@pytest.mark.parametrize('feed_type', ['rss', 'atom'])
+@pytest.mark.parametrize('feed_type', ['rss', 'atom', 'json'])
 def test_parse_returns_etag_last_modified(
     monkeypatch,
     parse,
@@ -428,13 +491,13 @@ def test_feedparser_parse_call(monkeypatch, parse, make_url, data_dir, exc_cls):
     assert feedparser_parse.kwargs['sanitize_html'] == True
 
 
-def test_missing_entry_id(parse):
+def test_missing_entry_id():
     """Handle RSS entries without guid.
     https://github.com/lemon24/reader/issues/170
 
     """
     # For RSS, when id is missing, parse() falls back to link.
-    feed, entries = parse_feed(
+    feed, entries = feedparser_parse(
         'url',
         """
         <?xml version="1.0" encoding="UTF-8" ?>
@@ -455,7 +518,7 @@ def test_missing_entry_id(parse):
 
     # ... and only link.
     with pytest.raises(ParseError) as excinfo:
-        parse_feed(
+        feedparser_parse(
             'url',
             """
             <?xml version="1.0" encoding="UTF-8" ?>
@@ -475,7 +538,7 @@ def test_missing_entry_id(parse):
 
     # There is no fallback for Atom.
     with pytest.raises(ParseError) as excinfo:
-        parse_feed(
+        feedparser_parse(
             'url',
             """
             <?xml version="1.0" encoding="utf-8"?>
@@ -492,14 +555,46 @@ def test_missing_entry_id(parse):
     assert excinfo.value.url == 'url'
     assert 'entry with no id' in excinfo.value.message
 
+    # Same for JSON Feed.
+    with pytest.raises(ParseError) as excinfo:
+        jsonfeed_parse(
+            'url',
+            """
+            {
+                "version": "https://jsonfeed.org/version/1.1",
+                "items": [
+                    {"content_text": "content"}
+                ]
+            }
+            """,
+        )
+    assert excinfo.value.url == 'url'
+    assert 'entry with no id' in excinfo.value.message
 
-def test_no_version(parse):
+    # Non-string, non-number id is treated as missing for JSON Feed.
+    with pytest.raises(ParseError) as excinfo:
+        jsonfeed_parse(
+            'url',
+            """
+            {
+                "version": "https://jsonfeed.org/version/1.1",
+                "items": [
+                    {"id": {"not": "a string"}, "content_text": "content"}
+                ]
+            }
+            """,
+        )
+    assert excinfo.value.url == 'url'
+    assert 'entry with no id' in excinfo.value.message
+
+
+def test_no_version():
     """Raise ParseError if feedparser can't detect the feed type and
     there's no bozo_exception.
 
     """
     with pytest.raises(ParseError) as excinfo:
-        parse_feed(
+        feedparser_parse(
             'url',
             """
             <?xml version="1.0" encoding="utf-8"?>
@@ -509,6 +604,29 @@ def test_no_version(parse):
     assert excinfo.value.url == 'url'
     assert 'unknown feed type' in excinfo.value.message
 
+    with pytest.raises(ParseError) as excinfo:
+        jsonfeed_parse(
+            'url',
+            """
+            {
+                "version": "https://bad.version/",
+                "items": [
+                    {"id": "1", "content_text": "content"}
+                ]
+            }
+            """,
+        )
+    assert excinfo.value.url == 'url'
+    assert 'missing or bad JSON Feed version' in excinfo.value.message
+
+
+def test_jsonfeed_invalid_json():
+    with pytest.raises(ParseError) as excinfo:
+        jsonfeed_parse('url', "malformed JSON")
+    assert excinfo.value.url == 'url'
+    assert 'invalid JSON' in excinfo.value.message
+    assert isinstance(excinfo.value.__cause__, json.JSONDecodeError)
+
 
 @pytest.fixture
 def make_http_set_headers_url(requests_mock):
@@ -550,30 +668,6 @@ def test_default_response_headers(
     parse(feed_url)
 
     assert mock.call_args[1]['response_headers']['Content-Location'] == feed_url
-    assert mock.call_args[1]['response_headers']['Content-Type'] == 'text/xml'
-
-
-def test_parsers(parse):
-    parse.parsers.clear()
-
-    parse.mount_parser('http://', lambda *args: ('generic', *args))
-    parse.mount_parser('http://specific.com', lambda *args: ('specific', *args))
-
-    assert parse('http://generic.com/', 'etag', None) == (
-        'generic',
-        'http://generic.com/',
-        'etag',
-        None,
-    )
-    assert parse('http://specific.com/', None, 'last_modified') == (
-        'specific',
-        'http://specific.com/',
-        None,
-        'last_modified',
-    )
-
-    with pytest.raises(ParseError):
-        parse('file:unknown')
 
 
 @pytest.mark.parametrize('scheme', ['', 'file:', 'file:///', 'file://localhost/'])
@@ -621,7 +715,7 @@ def test_feed_root_none(data_dir, scheme):
     with pytest.raises(ParseError) as excinfo:
         parse(url)
     assert excinfo.value.url == url
-    assert 'no parser' in excinfo.value.message
+    assert 'no retriever' in excinfo.value.message
 
 
 @pytest.mark.parametrize('scheme', ['', 'file:'])
@@ -743,8 +837,151 @@ def test_normalize_url_errors(monkeypatch, reload_module, os_name, url, reason):
     reload_module(urllib.request)
 
     with pytest.raises(ValueError) as excinfo:
-        FileParser(data_dir)._normalize_url(url)
+        FileRetriever(data_dir)._normalize_url(url)
 
     reload_module.undo()
 
     assert reason in str(excinfo.value)
+
+
+def test_parser_mount_order():
+    parse = Parser()
+    parse.mount_parser_by_mime_type('P0', 'one/two;q=0.0')
+    parse.mount_parser_by_mime_type('P1', 'one/two')
+    parse.mount_parser_by_mime_type('P2', 'one/two;q=0.1')
+    parse.mount_parser_by_mime_type('P3', 'one/two;q=0.1')
+    parse.mount_parser_by_mime_type('P4', 'one/two;q=0.4')
+    parse.mount_parser_by_mime_type('P5', 'one/two;q=0.5')
+    parse.mount_parser_by_mime_type('P6', 'one/two;q=0.3')
+    assert parse.parsers_by_mime_type == {
+        'one/two': [
+            (0.1, 'P2'),
+            (0.1, 'P3'),
+            (0.3, 'P6'),
+            (0.4, 'P4'),
+            (0.5, 'P5'),
+            (1, 'P1'),
+        ]
+    }
+
+
+def make_dummy_retriever(name, mime_type='type/subtype', headers=None):
+    @contextmanager
+    def retriever(url, http_etag, http_last_modified, http_accept):
+        retriever.last_http_accept = http_accept
+        yield RetrieveResult(name, mime_type, http_etag, http_last_modified, headers)
+
+    return retriever
+
+
+def make_dummy_parser(prefix='', http_accept=None):
+    def parser(url, file, headers):
+        parser.last_headers = headers
+        return prefix + file, url
+
+    if http_accept:
+        parser.http_accept = http_accept
+
+    return parser
+
+
+def test_parser_selection():
+    parse = Parser()
+
+    http_retriever = make_dummy_retriever('http', 'type/http', 'headers')
+    parse.mount_retriever('http:', http_retriever)
+    file_retriever = make_dummy_retriever('file', 'type/file')
+    parse.mount_retriever('file:', file_retriever)
+    nomt_retriever = make_dummy_retriever('nomt', None)
+    parse.mount_retriever('nomt:', nomt_retriever)
+    parse.mount_retriever('unkn:', make_dummy_retriever('unkn', 'type/unknown'))
+
+    http_parser = make_dummy_parser('httpp-', 'type/http')
+    parse.mount_parser_by_mime_type(http_parser)
+    assert parse('http:one', 'etag', None) == (
+        'httpp-http',
+        'http:one',
+        'etag',
+        None,
+    )
+    assert http_retriever.last_http_accept == 'type/http'
+    assert http_parser.last_headers == 'headers'
+
+    # this should not get in the way of anything else;
+    # it's mounted with "o" to check we do exact match, and not prefix match
+    url_parser = make_dummy_parser('urlp-')
+    parse.mount_parser_by_url('file:o', url_parser)
+
+    with pytest.raises(ParseError) as excinfo:
+        parse('file:one', None, 'last-modified')
+    assert excinfo.value.url == 'file:one'
+    assert 'no parser for MIME type' in excinfo.value.message
+    assert 'type/file' in excinfo.value.message
+
+    file_parser = make_dummy_parser('filep-')
+    parse.mount_parser_by_mime_type(file_parser, 'type/file, text/plain;q=0.8')
+    assert parse('file:one', None, 'last-modified') == (
+        'filep-file',
+        'file:one',
+        None,
+        'last-modified',
+    )
+    assert file_retriever.last_http_accept == 'type/http,type/file,text/plain;q=0.8'
+    assert file_parser.last_headers is None
+
+    with pytest.raises(ParseError) as excinfo:
+        parse('nomt:one')
+    assert excinfo.value.url == 'nomt:one'
+    assert 'no parser for MIME type' in excinfo.value.message
+    assert 'application/octet-stream' in excinfo.value.message
+
+    with pytest.raises(ParseError) as excinfo:
+        parse('nomt:one.html')
+    assert excinfo.value.url == 'nomt:one.html'
+    assert 'no parser for MIME type' in excinfo.value.message
+    assert 'text/html' in excinfo.value.message
+
+    with pytest.raises(ParseError) as excinfo:
+        parse('unkn:one')
+    assert excinfo.value.url == 'unkn:one'
+    assert 'no parser for MIME type' in excinfo.value.message
+    assert 'type/unknown' in excinfo.value.message
+
+    with pytest.raises(TypeError) as excinfo:
+        parse.mount_parser_by_mime_type(make_dummy_parser('fallbackp-'))
+    assert "unaware parser" in str(excinfo.value)
+
+    parse.mount_parser_by_mime_type(make_dummy_parser('fallbackp-'), '*/*')
+    assert parse('nomt:one') == ('fallbackp-nomt', 'nomt:one', None, None)
+    assert parse('unkn:one') == ('fallbackp-unkn', 'unkn:one', None, None)
+    assert nomt_retriever.last_http_accept == 'type/http,type/file,*/*,text/plain;q=0.8'
+
+    assert parse('file:o') == ('urlp-file', 'file:o', None, None)
+    assert file_retriever.last_http_accept is None
+    assert parse('file:///o') == ('urlp-file', 'file:///o', None, None)
+
+
+def test_retriever_selection():
+    parse = Parser()
+
+    parse.mount_retriever('http://', make_dummy_retriever('generic'))
+    parse.mount_retriever('http://specific.com', make_dummy_retriever('specific'))
+    parse.mount_parser_by_mime_type(make_dummy_parser(), '*/*')
+
+    assert parse('http://generic.com/', 'etag', None) == (
+        'generic',
+        'http://generic.com/',
+        'etag',
+        None,
+    )
+    assert parse('http://specific.com/', None, 'last-modified') == (
+        'specific',
+        'http://specific.com/',
+        None,
+        'last-modified',
+    )
+
+    with pytest.raises(ParseError) as excinfo:
+        parse('file:unknown')
+    assert excinfo.value.url == 'file:unknown'
+    assert 'no retriever' in excinfo.value.message
