@@ -17,8 +17,11 @@ from utils import rename_argument
 from reader import Content
 from reader import Enclosure
 from reader import Entry
+from reader import EntryCounts
 from reader import EntryNotFoundError
+from reader import EntrySearchCounts
 from reader import Feed
+from reader import FeedCounts
 from reader import FeedExistsError
 from reader import FeedNotFoundError
 from reader import make_reader
@@ -502,7 +505,12 @@ def raises_not_modified_parser(_):
 
 
 @pytest.mark.parametrize(
-    'make_new_parser', [same_parser, updated_feeds_parser, raises_not_modified_parser,]
+    'make_new_parser',
+    [
+        same_parser,
+        updated_feeds_parser,
+        raises_not_modified_parser,
+    ],
 )
 @rename_argument('reader', 'reader_with_one_feed')
 def test_last_exception_reset(reader, call_update_method, make_new_parser):
@@ -559,7 +567,10 @@ class EntryAction(Enum):
         (f, e)
         for f in FeedAction
         for e in EntryAction
-        if (f, e) not in {(FeedAction.none, EntryAction.none),}
+        if (f, e)
+        not in {
+            (FeedAction.none, EntryAction.none),
+        }
     ],
 )
 def test_update_feed_deleted(db_path, call_update_method, feed_action, entry_action):
@@ -727,7 +738,14 @@ class FakeNow:
         return self.now
 
 
-def get_entries_default(reader, **kwargs):
+# TODO: we could just parametrize sort as well here (see test_pagination_basic for details)
+
+
+def get_feeds(reader, **kwargs):
+    return reader.get_feeds(**kwargs)
+
+
+def get_entries(reader, **kwargs):
     return reader.get_entries(**kwargs)
 
 
@@ -744,7 +762,7 @@ def enable_and_update_search(reader):
     reader.update_search()
 
 
-def search_entries_default(reader, **kwargs):
+def search_entries(reader, **kwargs):
     return reader.search_entries('entry', **kwargs)
 
 
@@ -756,10 +774,14 @@ def search_entries_recent(reader, **kwargs):
     return reader.search_entries('entry', sort='recent', **kwargs)
 
 
+def search_entries_random(reader, **kwargs):
+    return reader.search_entries('entry', sort='random', **kwargs)
+
+
 with_call_entries_recent_method = pytest.mark.parametrize(
     'pre_stuff, call_method',
     [
-        (lambda _: None, get_entries_default),
+        (lambda _: None, get_entries),
         (lambda _: None, get_entries_recent),
         (enable_and_update_search, search_entries_recent),
     ],
@@ -957,7 +979,14 @@ def test_get_entries_recent_feed_order(reader, chunk_size, pre_stuff, call_metho
 
 
 @pytest.mark.parametrize('chunk_size', [1, 2, 3, 4])
-def test_get_entries_random(reader, chunk_size):
+@pytest.mark.parametrize(
+    'pre_stuff, call_method',
+    [
+        (lambda _: None, get_entries),
+        (enable_and_update_search, search_entries),
+    ],
+)
+def test_get_entries_random(reader, chunk_size, pre_stuff, call_method):
     """Black box get_entries(sort='random') good enough™ test.
 
     To have a more open-box test we'd need to:
@@ -987,13 +1016,15 @@ def test_get_entries_random(reader, chunk_size):
     reader.add_feed(feed.url)
     reader.update_feeds()
 
+    pre_stuff(reader)
+
     # all possible get_entries(sort='random') results
     all_tuples = set(permutations({e.id for e in reader.get_entries()}, chunk_size))
 
     # some get_entries(sort='random') results
     # (we call it enough times so it's likely we get all the results)
     random_tuples = Counter(
-        tuple(e.id for e in reader.get_entries(sort='random'))
+        tuple(e.id for e in call_method(reader, sort='random'))
         for _ in range(20 * len(all_tuples))
     )
 
@@ -1445,6 +1476,7 @@ with_call_entries_method = pytest.mark.parametrize(
         (lambda _: None, get_entries_random),
         (enable_and_update_search, search_entries_relevant),
         (enable_and_update_search, search_entries_recent),
+        (enable_and_update_search, search_entries_random),
     ],
 )
 
@@ -1481,7 +1513,16 @@ ALL_IDS = {
         (dict(entry=None), ALL_IDS),
         (dict(entry=('1', '1, 1')), {(1, 1)}),
         (dict(entry=('1', '1, 2')), {(1, 2)}),
-        (dict(entry=Entry('1, 2', datetime(2010, 2, 1), feed=Feed('1'),)), {(1, 2)},),
+        (
+            dict(
+                entry=Entry(
+                    '1, 2',
+                    datetime(2010, 2, 1),
+                    feed=Feed('1'),
+                )
+            ),
+            {(1, 2)},
+        ),
         (dict(entry=('inexistent', 'also-inexistent')), set()),
     ],
 )
@@ -1680,6 +1721,13 @@ with_call_feed_tags_method = pytest.mark.parametrize(
             get_entry_id,
             noop,
         ),
+        (
+            enable_and_update_search,
+            search_entries_random,
+            'feed_tags',
+            get_entry_id,
+            noop,
+        ),
         # TODO: maybe test all the get_feeds sort orders
         (lambda _: None, get_feeds, 'tags', get_feed_url, lambda t: t[0]),
     ],
@@ -1774,7 +1822,7 @@ def test_filtering_tags(
     assert actual_set == expected_set, kwargs
 
 
-ALL_IDS = {1, 2, 3}
+ALL_IDS = {1, 2, 3, 4}
 
 
 @pytest.mark.parametrize(
@@ -1786,6 +1834,9 @@ ALL_IDS = {1, 2, 3}
         (dict(broken=None), ALL_IDS),
         (dict(broken=True), {2}),
         (dict(broken=False), ALL_IDS - {2}),
+        (dict(updates_enabled=None), ALL_IDS),
+        (dict(updates_enabled=True), ALL_IDS - {4}),
+        (dict(updates_enabled=False), {4}),
     ],
 )
 def test_feeds_filtering(reader, kwargs, expected):
@@ -1794,10 +1845,13 @@ def test_feeds_filtering(reader, kwargs, expected):
     one = parser.feed(1, datetime(2010, 1, 1))
     two = parser.feed(2, datetime(2010, 1, 1))
     three = parser.feed(3, datetime(2010, 1, 1))
+    four = parser.feed(4, datetime(2010, 1, 1))
 
-    for feed in one, two, three:
+    for feed in one, two, three, four:
         reader.add_feed(feed)
 
+    reader.disable_feed_updates(four)
+
     reader.update_feeds()
 
     assert {eval(f.url) for f in reader.get_feeds(**kwargs)} == expected
@@ -1806,7 +1860,12 @@ def test_feeds_filtering(reader, kwargs, expected):
 
 
 @pytest.mark.parametrize(
-    'kwargs', [dict(feed=object()), dict(broken=object()),],
+    'kwargs',
+    [
+        dict(feed=object()),
+        dict(broken=object()),
+        dict(updates_enabled=object()),
+    ],
 )
 def test_feeds_filtering_error(reader, kwargs):
     reader._parser = parser = Parser()
@@ -1870,7 +1929,10 @@ def test_change_feed_url_feed(reader):
 
     assert not reader.get_feed('1', None)
     assert reader.get_feed('3') == old_one._replace(
-        url='3', updated=None, last_updated=None, last_exception=None,
+        url='3',
+        updated=None,
+        last_updated=None,
+        last_exception=None,
     )
 
 
@@ -1946,7 +2008,9 @@ def test_change_feed_url_second_update(reader, new_feed_url):
     reader.change_feed_url('1', new_feed_url)
 
     assert reader.get_feed(new_feed_url) == old_one._replace(
-        url=new_feed_url, updated=None, last_updated=None,
+        url=new_feed_url,
+        updated=None,
+        last_updated=None,
     )
 
     reader._parser.feed(
@@ -2108,3 +2172,387 @@ def test_change_feed_url_tags(reader):
 
 
 # END change_feed_url tests
+
+
+def test_updates_enabled(reader):
+    """Test Feed.updates_enabled functionality.
+
+    https://github.com/lemon24/reader/issues/187#issuecomment-706539658
+
+    """
+    reader._parser = parser = FailingParser(condition=lambda url: False)
+
+    one = parser.feed(1, datetime(2010, 1, 1))
+    one_one = parser.entry(1, 1, datetime(2010, 1, 1))
+    two = parser.feed(2, datetime(2010, 1, 1))
+    two_one = parser.entry(2, 1, datetime(2010, 1, 1))
+
+    for feed in one, two:
+        reader.add_feed(feed)
+
+    # updates_enabled is true by default
+    assert all(f.updates_enabled is True for f in reader.get_feeds())
+
+    # sanity check that feeds were not updated yet
+    assert all(f.updated is None for f in reader.get_feeds())
+    assert {e.id for e in reader.get_entries()} == set()
+
+    # sanity check that feeds update normally
+    reader.update_feeds()
+    assert all(f.updated == datetime(2010, 1, 1) for f in reader.get_feeds())
+    assert {e.id: e.updated for e in reader.get_entries()} == {
+        one_one.id: datetime(2010, 1, 1),
+        two_one.id: datetime(2010, 1, 1),
+    }
+
+    # make one feed fail temporarily, so last_exception gets set
+    parser.condition = lambda url: url == one.url
+    reader.update_feeds()
+    last_exception = reader.get_feed(one).last_exception
+    assert last_exception is not None
+    parser.condition = lambda url: False
+
+    # disable_feed_updates sets updates_enabled to False
+    reader.disable_feed_updates(one)
+    assert reader.get_feed(one).updates_enabled is False
+    # disable_feed_updates does not clear last_exception
+    assert reader.get_feed(one).last_exception == last_exception
+    # disable_feed_updates can be called twice
+    reader.disable_feed_updates(one)
+
+    one = parser.feed(1, datetime(2010, 1, 2))
+    one_one = parser.entry(1, 1, datetime(2010, 1, 2))
+    two = parser.feed(2, datetime(2010, 1, 2))
+    two_one = parser.entry(2, 1, datetime(2010, 1, 2))
+
+    # update_feeds skips feeds with updates_enabled == False
+    reader.update_feeds()
+    assert {f.url: f.updated for f in reader.get_feeds()} == {
+        one.url: datetime(2010, 1, 1),
+        two.url: datetime(2010, 1, 2),
+    }
+    assert {e.id: e.updated for e in reader.get_entries()} == {
+        one_one.id: datetime(2010, 1, 1),
+        two_one.id: datetime(2010, 1, 2),
+    }
+
+    # update_feed does not care about updates_enabled
+    reader.update_feed(one)
+    assert reader.get_feed(one).updated == datetime(2010, 1, 2)
+    assert reader.get_entry(one_one).updated == datetime(2010, 1, 2)
+
+    one = parser.feed(1, datetime(2010, 1, 3))
+    one_one = parser.entry(1, 1, datetime(2010, 1, 3))
+    two = parser.feed(2, datetime(2010, 1, 3))
+    two_one = parser.entry(2, 1, datetime(2010, 1, 3))
+
+    # update_feeds skips feeds with updates_enabled == False (again)
+    reader.update_feeds()
+    assert {f.url: f.updated for f in reader.get_feeds()} == {
+        one.url: datetime(2010, 1, 2),
+        two.url: datetime(2010, 1, 3),
+    }
+    assert {e.id: e.updated for e in reader.get_entries()} == {
+        one_one.id: datetime(2010, 1, 2),
+        two_one.id: datetime(2010, 1, 3),
+    }
+
+    # enable_feed_updates sets updates_enabled to True
+    reader.enable_feed_updates(one)
+    assert reader.get_feed(one).updates_enabled is True
+    # enable_feed_updates can be called twice
+    reader.enable_feed_updates(one)
+
+    # update_feeds updates newly-enabled feeds
+    reader.update_feeds()
+    assert reader.get_feed(one).updated == datetime(2010, 1, 3)
+    assert reader.get_entry(one_one).updated == datetime(2010, 1, 3)
+
+
+def test_updates_enabled_errors(reader):
+    with pytest.raises(FeedNotFoundError) as excinfo:
+        reader.enable_feed_updates('one')
+    assert excinfo.value.url == 'one'
+    assert 'no such feed' in excinfo.value.message
+
+    with pytest.raises(FeedNotFoundError) as excinfo:
+        reader.disable_feed_updates('one')
+    assert excinfo.value.url == 'one'
+    assert 'no such feed' in excinfo.value.message
+
+
+@pytest.mark.parametrize(
+    'kwargs, expected',
+    [
+        (dict(), FeedCounts(3, broken=1, updates_enabled=2)),
+        (dict(feed='1'), FeedCounts(1, 0, 1)),
+        (dict(tags=['tag']), FeedCounts(2, broken=1, updates_enabled=2)),
+        (dict(broken=True), FeedCounts(1, broken=1, updates_enabled=1)),
+        (dict(broken=False), FeedCounts(2, broken=0, updates_enabled=1)),
+        (dict(updates_enabled=True), FeedCounts(2, broken=1, updates_enabled=2)),
+        (dict(updates_enabled=False), FeedCounts(1, broken=0, updates_enabled=0)),
+        (dict(broken=True, updates_enabled=False), FeedCounts(0, 0, 0)),
+    ],
+)
+def test_feed_counts(reader, kwargs, expected):
+    # TODO: fuzz get_feeds() == get_feed_counts()
+
+    reader._parser = parser = FailingParser(condition=lambda url: False)
+
+    one = parser.feed(1, datetime(2010, 1, 1))
+    two = parser.feed(2, datetime(2010, 1, 1))
+    three = parser.feed(3, datetime(2010, 1, 1))
+
+    for feed in one, two, three:
+        reader.add_feed(feed)
+
+    parser.condition = lambda url: url == two.url
+    reader.disable_feed_updates(three)
+    reader.add_feed_tag(one, 'tag')
+    reader.add_feed_tag(two, 'tag')
+
+    reader.update_feeds()
+
+    assert reader.get_feed_counts(**kwargs) == expected
+
+
+def get_entry_counts(reader, **kwargs):
+    return reader.get_entry_counts(**kwargs)
+
+
+def search_entry_counts(reader, **kwargs):
+    return reader.search_entry_counts('entry', **kwargs)
+
+
+with_call_entry_counts_method = pytest.mark.parametrize(
+    'pre_stuff, call_method, rv_type',
+    [
+        (lambda _: None, get_entry_counts, EntryCounts),
+        (enable_and_update_search, search_entry_counts, EntrySearchCounts),
+    ],
+)
+
+
+@pytest.mark.parametrize(
+    'kwargs, expected',
+    [
+        (dict(), EntryCounts(9, read=2, important=4, has_enclosures=8)),
+        (dict(feed='1'), EntryCounts(1, read=0, important=0, has_enclosures=0)),
+        (dict(feed='2'), EntryCounts(8, read=2, important=4, has_enclosures=8)),
+        (
+            dict(entry=('1', '1, 1')),
+            EntryCounts(1, read=0, important=0, has_enclosures=0),
+        ),
+        (
+            dict(entry=('2', '2, 1')),
+            EntryCounts(1, read=1, important=1, has_enclosures=1),
+        ),
+        (
+            dict(entry=('2', '2, 3')),
+            EntryCounts(1, read=0, important=1, has_enclosures=1),
+        ),
+        (
+            dict(entry=('2', '2, 5')),
+            EntryCounts(1, read=0, important=0, has_enclosures=1),
+        ),
+        (dict(read=True), EntryCounts(2, read=2, important=2, has_enclosures=2)),
+        (dict(read=False), EntryCounts(7, read=0, important=2, has_enclosures=6)),
+        (dict(important=True), EntryCounts(4, read=2, important=4, has_enclosures=4)),
+        (dict(important=False), EntryCounts(5, read=0, important=0, has_enclosures=4)),
+        (
+            dict(has_enclosures=True),
+            EntryCounts(8, read=2, important=4, has_enclosures=8),
+        ),
+        (
+            dict(has_enclosures=False),
+            EntryCounts(1, read=0, important=0, has_enclosures=0),
+        ),
+        (
+            dict(feed_tags=['tag']),
+            EntryCounts(1, read=0, important=0, has_enclosures=0),
+        ),
+    ],
+)
+@with_call_entry_counts_method
+def test_entry_counts(reader, kwargs, expected, pre_stuff, call_method, rv_type):
+    # TODO: fuzz get_entries() == get_entry_counts()
+
+    reader._parser = parser = Parser()
+
+    one = parser.feed(1, datetime(2010, 1, 3))
+    two = parser.feed(2, datetime(2010, 1, 3))
+    three = parser.feed(3, datetime(2010, 1, 1))
+    one_entry = parser.entry(
+        1,
+        1,
+        datetime(2010, 1, 3),
+        summary='summary',
+        content=(Content('value3', 'type', 'en'), Content('value2')),
+    )
+    two_entries = [
+        parser.entry(2, i, datetime(2010, 1, 3), enclosures=[]) for i in range(1, 1 + 8)
+    ]
+
+    # TODO: less overlap would be nice (e.g. some read that don't have enclosures)
+
+    for entry in two_entries[:8]:
+        int_feed_url, int_id = eval(entry.id)
+        parser.entries[int_feed_url][int_id].enclosures.append(Enclosure('http://e'))
+
+    for feed in one, two, three:
+        reader.add_feed(feed)
+
+    reader.add_feed_tag(one, 'tag')
+
+    reader.update_feeds()
+    pre_stuff(reader)
+
+    for entry in two_entries[:2]:
+        reader.mark_as_read(entry)
+    for entry in two_entries[:4]:
+        reader.mark_as_important(entry)
+
+    rv = call_method(reader, **kwargs)
+    assert type(rv) is rv_type
+    # this isn't gonna work as well if the return types get different attributes
+    assert rv._asdict() == expected._asdict()
+
+
+with_call_paginated_method = pytest.mark.parametrize(
+    'pre_stuff, call_method, sort_kwargs',
+    [
+        (lambda _: None, get_feeds, {}),
+        (lambda _: None, get_feeds, dict(sort='title')),
+        (lambda _: None, get_feeds, dict(sort='added')),
+        (lambda _: None, get_entries, {}),
+        (lambda _: None, get_entries, dict(sort='recent')),
+        (enable_and_update_search, search_entries, {}),
+        (enable_and_update_search, search_entries, dict(sort='relevant')),
+        (enable_and_update_search, search_entries, dict(sort='recent')),
+    ],
+)
+
+
+@pytest.fixture
+def reader_with_three_feeds(reader):
+    reader._parser = parser = Parser()
+
+    one = parser.feed(1, datetime(2010, 1, 1))
+    parser.entry(1, 1, datetime(2010, 1, 1))
+    parser.entry(1, 2, datetime(2010, 1, 1))
+    parser.entry(1, 3, datetime(2010, 1, 1))
+    two = parser.feed(2, datetime(2010, 1, 1))
+    three = parser.feed(3, datetime(2010, 1, 1))
+
+    for feed in one, two, three:
+        reader.add_feed(feed)
+
+    return reader
+
+
+@with_call_paginated_method
+@pytest.mark.parametrize('chunk_size', [Storage.chunk_size, 0, 1, 2])
+@rename_argument('reader', 'reader_with_three_feeds')
+def test_pagination_basic(reader, pre_stuff, call_method, sort_kwargs, chunk_size):
+    reader._storage.chunk_size = chunk_size
+
+    reader.update_feeds()
+    pre_stuff(reader)
+
+    def get_ids(**kwargs):
+        return [o.object_id for o in call_method(reader, **sort_kwargs, **kwargs)]
+
+    ids = get_ids()
+
+    assert get_ids(starting_after=ids[0]) == ids[1:]
+    assert get_ids(starting_after=ids[1]) == ids[2:]
+    assert get_ids(starting_after=ids[2]) == ids[3:] == []
+
+    assert get_ids(limit=1) == ids[:1]
+    assert get_ids(limit=2) == ids[:2]
+    assert get_ids(limit=3) == ids[:3] == ids
+
+    assert get_ids(limit=1, starting_after=ids[0]) == ids[1:2]
+    assert get_ids(limit=2, starting_after=ids[0]) == ids[1:3]
+    assert get_ids(limit=2, starting_after=ids[1]) == ids[2:]
+    assert get_ids(limit=2, starting_after=ids[2]) == ids[3:] == []
+
+
+@pytest.mark.parametrize(
+    'pre_stuff, call_method',
+    [
+        (lambda _: None, get_entries),
+        (enable_and_update_search, search_entries),
+    ],
+)
+@pytest.mark.parametrize('chunk_size', [Storage.chunk_size, 0, 1, 2])
+@rename_argument('reader', 'reader_with_three_feeds')
+def test_pagination_random(reader, pre_stuff, call_method, chunk_size):
+    reader._storage.chunk_size = chunk_size
+
+    reader.update_feeds()
+    pre_stuff(reader)
+
+    def get_ids(**kwargs):
+        return [o.object_id for o in call_method(reader, sort='random', **kwargs)]
+
+    ids = [o.object_id for o in call_method(reader)]
+
+    assert len(get_ids(limit=1)) == min(1, chunk_size or 1, len(ids))
+    assert len(get_ids(limit=2)) == min(2, chunk_size or 2, len(ids))
+    assert len(get_ids(limit=3)) == min(3, chunk_size or 3, len(ids))
+
+    with pytest.raises(ValueError):
+        get_ids(starting_after=ids[0])
+
+    with pytest.raises(ValueError):
+        get_ids(limit=1, starting_after=ids[0])
+
+
+NOT_FOUND_ERROR_CLS = {
+    get_feeds: FeedNotFoundError,
+    get_entries: EntryNotFoundError,
+    search_entries: EntryNotFoundError,
+}
+
+NOT_FOUND_STARTING_AFTER = {
+    get_feeds: '0',
+    get_entries: ('1', '1, 0'),
+    search_entries: ('1', '1, 0'),
+}
+
+
+@with_call_paginated_method
+def test_starting_after_errors(reader, pre_stuff, call_method, sort_kwargs):
+    pre_stuff(reader)
+
+    error_cls = NOT_FOUND_ERROR_CLS[call_method]
+    starting_after = NOT_FOUND_STARTING_AFTER[call_method]
+
+    with pytest.raises(error_cls) as excinfo:
+        list(call_method(reader, **sort_kwargs, starting_after=starting_after))
+    assert excinfo.value.object_id == starting_after
+
+
+@with_call_paginated_method
+def test_limit_errors(reader, pre_stuff, call_method, sort_kwargs):
+    pre_stuff(reader)
+
+    def get_ids(**kwargs):
+        return [o.object_id for o in call_method(reader, **sort_kwargs, **kwargs)]
+
+    with pytest.raises(ValueError):
+        get_ids(limit=object())
+    with pytest.raises(ValueError):
+        get_ids(limit=0)
+    with pytest.raises(ValueError):
+        get_ids(limit=-1)
+    with pytest.raises(ValueError):
+        get_ids(limit=1.0)
+
+
+def test_logging_defaults():
+    logger = logging.getLogger('reader')
+    assert logger.level == logging.NOTSET
+    assert len(logger.handlers) == 1
+    assert isinstance(logger.handlers[0], logging.NullHandler)
