@@ -7,7 +7,8 @@ from fakeparser import ParserThatRemembers
 from reader import Entry
 from reader import Feed
 from reader import FeedNotFoundError
-from reader.core.reader import _missing
+from reader import make_reader
+from reader._storage import Storage
 
 
 def test_update_stale(reader, call_update_method):
@@ -28,20 +29,44 @@ def test_update_stale(reader, call_update_method):
 
     reader.add_feed(feed.url)
 
+    reader._now = lambda: datetime(2010, 1, 1)
     call_update_method(reader, feed.url)
-    assert set(reader.get_entries()) == {entry.as_entry(feed=feed)}
+
+    assert set((f.url, f.title, f.last_updated) for f in reader.get_feeds()) == {
+        (feed.url, feed.title, datetime(2010, 1, 1))
+    }
+    # FIXME: use entry.last_updated once we have it
+    assert set((e.id, e.title) for e in reader.get_entries()) == {
+        (entry.id, entry.title)
+    }
 
     new_feed = parser.feed(1, datetime(2010, 1, 1), title="new feed title")
     new_entry = parser.entry(1, 1, datetime(2010, 1, 1), title="new entry title")
 
+    # nothing changes after update
+    reader._now = lambda: datetime(2010, 1, 2)
     call_update_method(reader, feed.url)
-    assert set(reader.get_entries()) == {entry.as_entry(feed=feed)}
+    assert set((f.url, f.title, f.last_updated) for f in reader.get_feeds()) == {
+        (feed.url, feed.title, datetime(2010, 1, 1))
+    }
+    # FIXME: use entry.last_updated once we have it
+    assert set((e.id, e.title) for e in reader.get_entries()) == {
+        (entry.id, entry.title)
+    }
 
+    # but it does if we mark the feed as stale
     parser.calls[:] = []
     reader._storage.mark_as_stale(feed.url)
+    reader._now = lambda: datetime(2010, 1, 3)
     call_update_method(reader, feed.url)
     assert parser.calls == [(feed.url, None, None)]
-    assert set(reader.get_entries()) == {new_entry.as_entry(feed=new_feed)}
+    assert set((f.url, f.title, f.last_updated) for f in reader.get_feeds()) == {
+        (feed.url, new_feed.title, datetime(2010, 1, 3))
+    }
+    # FIXME: use entry.last_updated once we have it
+    assert set((e.id, e.title) for e in reader.get_entries()) == {
+        (entry.id, new_entry.title)
+    }
 
 
 def test_update_parse(reader, call_update_method):
@@ -85,6 +110,7 @@ def test_post_entry_add_plugins(reader):
     reader._post_entry_add_plugins.append(first_plugin)
     reader.update_feeds()
     assert plugin_calls == [(first_plugin, one)]
+    assert set(e.id for e in reader.get_entries()) == {'1, 1'}
 
     plugin_calls[:] = []
 
@@ -97,11 +123,12 @@ def test_post_entry_add_plugins(reader):
         (first_plugin, two),
         (second_plugin, two),
     ]
-    assert set(reader.get_entries()) == {
-        one.as_entry(feed=feed),
-        two.as_entry(feed=feed),
-    }
+    assert set(e.id for e in reader.get_entries()) == {'1, 1', '1, 2'}
+
+    # TODO: What is the expected behavior if a plugin raises an exception?
 
 
-def test_missing():
-    assert repr(_missing) == 'no value'
+def test_make_reader_storage():
+    storage = Storage(':memory:')
+    reader = make_reader('', _storage=storage)
+    assert reader._storage is storage
