@@ -125,6 +125,10 @@ def set_feed_user_title(storage, feed, __):
     storage.set_feed_user_title(feed.url, 'title')
 
 
+def set_feed_updates_enabled(storage, feed, __):
+    storage.set_feed_updates_enabled(feed.url, 1)
+
+
 def mark_as_stale(storage, feed, __):
     storage.mark_as_stale(feed.url)
 
@@ -185,6 +189,22 @@ def get_feed_tags(storage, feed, __):
     list(storage.get_feed_tags(feed.url))
 
 
+def get_feed_counts(storage, _, __):
+    storage.get_feed_counts()
+
+
+def get_entry_counts(storage, _, __):
+    storage.get_entry_counts(),
+
+
+def get_feed_last(storage, feed, __):
+    storage.get_feed_last('title', feed.url)
+
+
+def get_entry_last(storage, feed, entry):
+    storage.get_entry_last(datetime(2010, 1, 1), 'recent', (feed.url, entry.id))
+
+
 @pytest.mark.slow
 @pytest.mark.parametrize(
     'do_stuff',
@@ -196,6 +216,7 @@ def get_feed_tags(storage, feed, __):
         get_feeds_for_update,
         get_entries_for_update,
         set_feed_user_title,
+        set_feed_updates_enabled,
         mark_as_stale,
         mark_as_read_unread,
         update_feed,
@@ -210,6 +231,10 @@ def get_feed_tags(storage, feed, __):
         add_feed_tag,
         remove_feed_tag,
         get_feed_tags,
+        get_feed_counts,
+        get_entry_counts,
+        get_feed_last,
+        get_entry_last,
     ],
 )
 def test_errors_locked(db_path, do_stuff):
