@@ -8,8 +8,12 @@ from hypothesis import example
 from hypothesis import given
 
 from reader import Entry
+from reader import EntryError
 from reader import EntrySearchResult
 from reader import Feed
+from reader import FeedError
+from reader._types import EntryData
+from reader._types import FeedData
 from reader.types import _entry_argument
 from reader.types import _feed_argument
 from reader.types import _namedtuple_compat
@@ -63,6 +67,16 @@ def test__entry_argument():
         _entry_argument(('a', 'b', 'c'))
 
 
+def test_object_id():
+    assert Feed('url').object_id == 'url'
+    assert Entry('entry', 'updated', feed=Feed('url')).object_id == ('url', 'entry')
+    assert EntrySearchResult('url', 'entry').object_id == ('url', 'entry')
+    assert FeedData('url').object_id == 'url'
+    assert EntryData('url', 'entry', 'updated').object_id == ('url', 'entry')
+    assert FeedError('url').object_id == 'url'
+    assert EntryError('url', 'entry').object_id == ('url', 'entry')
+
+
 @pytest.mark.parametrize(
     'highlights',
     [
